--
-- use the pg_dump -t 'schema-name.table-name' --schema-only database-name
-- command to dump the creation sql scripts of the previous schema, and make changes based on that.
--
SET statement_timeout = 0;
SET lock_timeout = 0;
SET idle_in_transaction_session_timeout = 0;
SET client_encoding = 'UTF8';
SET standard_conforming_strings = on;
SELECT pg_catalog.set_config('search_path', '', false);
SET check_function_bodies = false;
SET xmloption = content;
SET client_min_messages = warning;
SET row_security = off;
SET default_tablespace = '';
SET default_table_access_method = heap;

--
-- Check if old table exists, if yes, do the data migration, if no, do nothing
--
CREATE OR REPLACE FUNCTION public.migrate_data(schema TEXT, old_table_name TEXT, new_table_name TEXT)
RETURNS BOOLEAN
AS
$$
DECLARE
    _tmp BOOLEAN;
    _columns TEXT;
BEGIN
    EXECUTE format('SELECT EXISTS(SELECT 1 FROM information_schema.tables WHERE
    table_schema=%L
    AND table_name=%L)', schema, old_table_name)
    INTO _tmp;
    IF _tmp THEN
        EXECUTE format('SELECT ARRAY_TO_STRING((SELECT array_agg(column_name::TEXT) FROM information_schema.columns WHERE  table_name=%L), %L)', old_table_name, ',')
        INTO _columns;
        RAISE INFO 'Columns  of old schema %', _columns;
        RAISE INFO 'Start to migrate existing data from % table to % table', old_table_name, new_table_name;
        EXECUTE format('INSERT INTO %I.%I(%s) SELECT %s FROM %I.%I', schema, new_table_name, _columns, _columns, schema, old_table_name);
        RAISE INFO 'Finish migrating existing data from % table to % table', old_table_name, new_table_name;
    END IF ;
    RETURN TRUE;
    END;
$$
LANGUAGE 'plpgsql';

--
-- Create table for both onprem and saas
--
CREATE OR REPLACE FUNCTION public.create_partition_table()
RETURNS BOOLEAN
AS
$$
BEGIN

    -- If there are now rows in dgs_sites table due to platform job failures, we cannot set the default value of site_id in recommendation table
    -- Hence fail the create psql job and let the caller retry by installing or upgrading the platform again , followed by nsxi
    IF NOT EXISTS (SELECT site_id FROM public.dgs_sites LIMIT 1) THEN
        RAISE INFO 'No rows in dgs_sites table, and hence we do not know the default site';
        RAISE INFO 'Hence failing the create_partition_table function';
        RAISE EXCEPTION 'Default site could not be determined since there are no rows in dgs_sites table'
              USING HINT = 'Please check that the load-default-site k8s job has run and succeeded during platform installation or upgrade';
    END IF;
    --
    -- for on-prem, we can't directly alter the existing table to be the partitioned table.
    -- so what we can do is to rename the existing one to be the old table
    -- and migrate the data from old table to the new partitioned table
    --
    ALTER TABLE IF EXISTS public.recommendation RENAME TO recommendation_old;

    --
    -- If old table exists, drop the old recommendation_pkey
    --
    ALTER TABLE IF EXISTS ONLY public.recommendation_old
    DROP CONSTRAINT IF EXISTS recommendation_pkey CASCADE;

    --
    -- create the partitioned recommendation table partitioned by site_id
    -- and add new column site_id
    --
    CREATE TABLE IF NOT EXISTS public.recommendation  (
        path character varying(255) NOT NULL,
        revision integer,
        createtime bigint,
        createuser character varying(255),
        description character varying(1024),
        displayname character varying(255),
        id character varying(255),
        lastmodifiedtime bigint,
        lastmodifieduser character varying(255),
        parentpath character varying(255),
        protection character varying(255),
        relativepath character varying(255),
        resourcetype character varying(255),
        systemowned boolean,
        connectivitypreference character varying(255) NOT NULL,
        connectivitypreferenceloggingenabled boolean,
        enforceapptiering boolean DEFAULT false,
        groupreusethreshold integer,
        grouptype character varying(255) NOT NULL,
        jobruntime bigint,
        manuallyrevised boolean,
        monitoringenabled boolean,
        permissivemode character varying(255) NOT NULL,
        policytype character varying(255) NOT NULL,
        status character varying(255) NOT NULL,
        statusmessage character varying(16384),
        endtime bigint NOT NULL,
        starttime bigint NOT NULL,
        additional_unmicrosegmented_rule_ids character varying(128)[] DEFAULT '{}'::character varying[],
        alert_reasons character varying(255)[] DEFAULT '{}'::character varying[],
        compute_exclusion_criteria character varying(255)[] DEFAULT '{}'::character varying[],
        consider_destination_any_rules_as_default boolean DEFAULT false,
        consider_service_any_rules_as_default boolean DEFAULT false,
        consider_source_any_rules_as_default boolean DEFAULT false,
        context_effective_physical_servers character varying(1024)[] DEFAULT '{}'::character varying[],
        context_effective_vms character varying(1024)[] DEFAULT '{}'::character varying[],
        context_groups character varying(1024)[] DEFAULT '{}'::character varying[],
        context_physical_servers character varying(1024)[] DEFAULT '{}'::character varying[],
        context_security_policies character varying(1024)[] DEFAULT '{}'::character varying[],
        context_vms character varying(1024)[] DEFAULT '{}'::character varying[],
        exclude_broadcast_flows boolean DEFAULT false,
        exclude_multicast_flows boolean DEFAULT false,
        ip_group_reuse_enabled boolean DEFAULT true,
        non_context_effective_physical_servers character varying(1024)[] DEFAULT '{}'::character varying[],
        non_context_effective_vms character varying(1024)[] DEFAULT '{}'::character varying[],
        infra jsonb,
        recommendation text,
        diff_document text,
        service_filter text,
        publish_infra jsonb,
        checkpoint jsonb,
        context_ips character varying(1024)[] DEFAULT '{}'::character varying[],
        ipv4_group_reuse_threshold integer DEFAULT 32,
        ipv6_group_reuse_threshold integer DEFAULT 128,
        site_id character varying(255) DEFAULT 'default',
        recommendationtype character varying(255) DEFAULT 'DFW',
        PRIMARY KEY (path, site_id)
    ) partition by hash(site_id);


    EXECUTE format('ALTER TABLE public.recommendation ALTER COLUMN site_id SET DEFAULT %L'
                 , (SELECT site_id FROM public.dgs_sites LIMIT 1));

    ALTER TABLE public.recommendation OWNER TO postgres;

    --
    -- create 10 hash partitions for recommendation table
    -- noted that the hash partition can't have default partition
    --
    create table public.recommendation_p0 partition of public.recommendation for values with (modulus 10, remainder 0);
    create table public.recommendation_p1 partition of public.recommendation for values with (modulus 10, remainder 1);
    create table public.recommendation_p2 partition of public.recommendation for values with (modulus 10, remainder 2);
    create table public.recommendation_p3 partition of public.recommendation for values with (modulus 10, remainder 3);
    create table public.recommendation_p4 partition of public.recommendation for values with (modulus 10, remainder 4);
    create table public.recommendation_p5 partition of public.recommendation for values with (modulus 10, remainder 5);
    create table public.recommendation_p6 partition of public.recommendation for values with (modulus 10, remainder 6);
    create table public.recommendation_p7 partition of public.recommendation for values with (modulus 10, remainder 7);
    create table public.recommendation_p8 partition of public.recommendation for values with (modulus 10, remainder 8);
    create table public.recommendation_p9 partition of public.recommendation for values with (modulus 10, remainder 9);

    --
    -- migrate the data
    --
    PERFORM public.migrate_data('public', 'recommendation_old', 'recommendation');

    --
    -- Adding index for rec table, that will create idx for all partitioned table automatically
    -- Need to create site_id idx explicitly since the pkey is composite key now.
    --
    CREATE INDEX rec_site_id_idx ON public.recommendation USING btree (site_id);

    CREATE INDEX rec_recommendation_type_idx ON public.recommendation USING btree(recommendationtype);

    DROP INDEX IF EXISTS public.rec_path_idx;
    CREATE INDEX rec_path_idx ON public.recommendation USING btree (path);

    DROP INDEX IF EXISTS public.rec_connect_preference_idx;
    CREATE INDEX rec_connect_preference_idx ON public.recommendation USING btree (connectivitypreference);

    DROP INDEX IF EXISTS public.rec_display_name_idx;
    CREATE INDEX rec_display_name_idx ON public.recommendation USING btree (displayname);

    DROP INDEX IF EXISTS public.rec_infra_idx;
    CREATE INDEX rec_infra_idx ON public.recommendation USING btree (infra);

    DROP INDEX IF EXISTS public.rec_monitor_enabled_idx;
    CREATE INDEX rec_monitor_enabled_idx ON public.recommendation USING btree (monitoringenabled);

    DROP INDEX IF EXISTS public.rec_permissive_mode_idx;
    CREATE INDEX rec_permissive_mode_idx ON public.recommendation USING btree (permissivemode);

    DROP INDEX IF EXISTS public.rec_status_idx;
    CREATE INDEX rec_status_idx ON public.recommendation USING btree (status);


    --
    -- Create the recommendation_monitor partition table
    --
    ALTER TABLE IF EXISTS public.recommendation_monitor
    RENAME TO recommendation_monitor_old;

    --
    -- Drop the old recommendation_monitor_pkey if exists
    --
    ALTER TABLE IF EXISTS ONLY public.recommendation_monitor_old
    DROP CONSTRAINT IF EXISTS recommendation_monitor_pkey CASCADE;

    --
    -- Create the recommendation_monitor partition table with new composite primary key
    --
    CREATE TABLE public.recommendation_monitor (
        path character varying(255) NOT NULL,
        lastmonitoringstarttime bigint,
        latest_effective_physical_servers character varying(1024)[] DEFAULT '{}'::character varying[],
        latest_effective_vms character varying(1024)[] DEFAULT '{}'::character varying[],
        latestpermissivemode character varying(255),
        rerunrecommended boolean NOT NULL,
        rerun_recommended_reasons character varying(1024)[] DEFAULT '{}'::character varying[],
        site_id character varying(255) DEFAULT 'default'
    ) partition by hash(site_id);

    EXECUTE format('ALTER TABLE public.recommendation_monitor ALTER COLUMN site_id SET DEFAULT %L'
                 , (SELECT site_id FROM public.dgs_sites LIMIT 1));

    ALTER TABLE public.recommendation_monitor OWNER TO postgres;

    ALTER TABLE ONLY public.recommendation_monitor
    ADD CONSTRAINT recommendation_monitor_pkey PRIMARY KEY (path, site_id);

    --
    -- Create 10 partitions for recommendation_monitor table
    --
    create table public.recommendation_monitor_p0 partition of public.recommendation_monitor for values with (modulus 10, remainder 0);
    create table public.recommendation_monitor_p1 partition of public.recommendation_monitor for values with (modulus 10, remainder 1);
    create table public.recommendation_monitor_p2 partition of public.recommendation_monitor for values with (modulus 10, remainder 2);
    create table public.recommendation_monitor_p3 partition of public.recommendation_monitor for values with (modulus 10, remainder 3);
    create table public.recommendation_monitor_p4 partition of public.recommendation_monitor for values with (modulus 10, remainder 4);
    create table public.recommendation_monitor_p5 partition of public.recommendation_monitor for values with (modulus 10, remainder 5);
    create table public.recommendation_monitor_p6 partition of public.recommendation_monitor for values with (modulus 10, remainder 6);
    create table public.recommendation_monitor_p7 partition of public.recommendation_monitor for values with (modulus 10, remainder 7);
    create table public.recommendation_monitor_p8 partition of public.recommendation_monitor for values with (modulus 10, remainder 8);
    create table public.recommendation_monitor_p9 partition of public.recommendation_monitor for values with (modulus 10, remainder 9);

    --
    -- Migrate the data from old monitor table to the new table
    PERFORM public.migrate_data('public', 'recommendation_monitor_old', 'recommendation_monitor');

    --
    -- Add all indexes for rec monitor table
    --
    DROP INDEX IF EXISTS public.rec_monitor_path_idx;
    CREATE INDEX rec_monitor_path_idx ON public.recommendation_monitor USING btree (path);

    DROP INDEX IF EXISTS public.rec_monitor_rerun_idx;
    CREATE INDEX rec_monitor_rerun_idx ON public.recommendation_monitor USING btree (rerunrecommended);

    CREATE INDEX rec_monitor_site_id_idx ON public.recommendation_monitor USING btree (site_id);

    --
    -- Add foreign key
    --
    ALTER TABLE public.recommendation_monitor
    ADD CONSTRAINT rec_monitor_foreign FOREIGN KEY (path, site_id) REFERENCES public.recommendation(path, site_id);

    --
    -- Create the recommendation_exclusion partition table
    --
    ALTER TABLE IF EXISTS public.recommendation_exclusion
    RENAME TO recommendation_exclusion_old;

    --
    -- Drop the old recommendation_exclusion_pkey if exists
    --
    ALTER TABLE IF EXISTS ONLY public.recommendation_exclusion_old
    DROP CONSTRAINT IF EXISTS recommendation_exclusion_pkey CASCADE;

    --
    -- Create recommendation exclusion partition table
    --
    CREATE TABLE public.recommendation_exclusion (
        exclusion_reason character varying(1024) NOT NULL,
        path character varying(1024) NOT NULL,
        excluded_context_physical_servers character varying(1024)[] DEFAULT '{}'::character varying[],
        excluded_context_vms character varying(1024)[] DEFAULT '{}'::character varying[],
        excluded_non_context_physical_servers character varying(1024)[] DEFAULT '{}'::character varying[],
        excluded_non_context_vms character varying(1024)[] DEFAULT '{}'::character varying[],
        recommendation_path character varying(255),
        site_id character varying(255) DEFAULT 'default',
        rec_site_id character varying(255) NOT NULL DEFAULT 'default'
    ) partition by hash(site_id);

    EXECUTE format('ALTER TABLE public.recommendation_exclusion ALTER COLUMN site_id SET DEFAULT %L'
                 , (SELECT site_id FROM public.dgs_sites LIMIT 1));
    EXECUTE format('ALTER TABLE public.recommendation_exclusion ALTER COLUMN rec_site_id SET DEFAULT %L'
                 , (SELECT site_id FROM public.dgs_sites LIMIT 1));

    ALTER TABLE public.recommendation_exclusion OWNER TO postgres;

    ALTER TABLE ONLY public.recommendation_exclusion
    ADD CONSTRAINT recommendation_exclusion_pkey PRIMARY KEY (exclusion_reason, path, site_id);

    --
    -- Create 10 partitions for recommendation_exclusion table
    --
    create table public.recommendation_exclusion_p0 partition of public.recommendation_exclusion for values with (modulus 10, remainder 0);
    create table public.recommendation_exclusion_p1 partition of public.recommendation_exclusion for values with (modulus 10, remainder 1);
    create table public.recommendation_exclusion_p2 partition of public.recommendation_exclusion for values with (modulus 10, remainder 2);
    create table public.recommendation_exclusion_p3 partition of public.recommendation_exclusion for values with (modulus 10, remainder 3);
    create table public.recommendation_exclusion_p4 partition of public.recommendation_exclusion for values with (modulus 10, remainder 4);
    create table public.recommendation_exclusion_p5 partition of public.recommendation_exclusion for values with (modulus 10, remainder 5);
    create table public.recommendation_exclusion_p6 partition of public.recommendation_exclusion for values with (modulus 10, remainder 6);
    create table public.recommendation_exclusion_p7 partition of public.recommendation_exclusion for values with (modulus 10, remainder 7);
    create table public.recommendation_exclusion_p8 partition of public.recommendation_exclusion for values with (modulus 10, remainder 8);
    create table public.recommendation_exclusion_p9 partition of public.recommendation_exclusion for values with (modulus 10, remainder 9);

    --
    -- migrate the date from old table to the new partition table
    --
    PERFORM public.migrate_data('public', 'recommendation_exclusion_old', 'recommendation_exclusion');

    --
    -- Add index for recommendation_exclusion table
    --
    DROP INDEX IF EXISTS public.rec_exclusion_path_idx;
    CREATE INDEX rec_exclusion_path_idx ON public.recommendation_exclusion USING btree (path);

    DROP INDEX IF EXISTS public.rec_exclusion_rec_path_idx;
    CREATE INDEX rec_exclusion_rec_path_idx ON public.recommendation_exclusion USING btree (recommendation_path);

    CREATE INDEX rec_exclusion_site_id_idx ON public.recommendation_exclusion USING btree (site_id);

    --
    -- Add foreign key
    --
    ALTER TABLE public.recommendation_exclusion
    ADD CONSTRAINT rec_exclusion_foreign FOREIGN KEY (recommendation_path, rec_site_id) REFERENCES public.recommendation(path, site_id);

    --
    -- create recommendation_partial_reused_group partition table
    --
    ALTER TABLE IF EXISTS public.recommendation_partial_reused_group
    RENAME TO recommendation_partial_reused_group_old;

    --
    -- Drop old pkey if exists
    --
    ALTER TABLE IF EXISTS ONLY public.recommendation_partial_reused_group_old
    DROP CONSTRAINT IF EXISTS recommendation_partial_reused_group_pkey CASCADE;

    --
    -- create the recommendation_partial_reused_group partition table
    --
    CREATE TABLE public.recommendation_partial_reused_group (
        path character varying(1024) NOT NULL,
        destination_groups text,
        source_groups text,
        recommendation_path character varying(255),
        site_id character varying(255) DEFAULT 'default',
        rec_site_id character varying(255) NOT NULL DEFAULT 'default'
    ) partition by hash(site_id);

    EXECUTE format('ALTER TABLE public.recommendation_partial_reused_group ALTER COLUMN site_id SET DEFAULT %L'
                 , (SELECT site_id FROM public.dgs_sites LIMIT 1));
    EXECUTE format('ALTER TABLE public.recommendation_partial_reused_group ALTER COLUMN rec_site_id SET DEFAULT %L'
                 , (SELECT site_id FROM public.dgs_sites LIMIT 1));


    ALTER TABLE public.recommendation_partial_reused_group OWNER TO postgres;

    ALTER TABLE ONLY public.recommendation_partial_reused_group
    ADD CONSTRAINT recommendation_partial_reused_group_pkey PRIMARY KEY (path, site_id);

    --
    -- create 10 partitions for recommendation_partial_reused_group
    --
    create table public.recommendation_partial_reused_group_p0 partition of public.recommendation_partial_reused_group for values with (modulus 10, remainder 0);
    create table public.recommendation_partial_reused_group_p1 partition of public.recommendation_partial_reused_group for values with (modulus 10, remainder 1);
    create table public.recommendation_partial_reused_group_p2 partition of public.recommendation_partial_reused_group for values with (modulus 10, remainder 2);
    create table public.recommendation_partial_reused_group_p3 partition of public.recommendation_partial_reused_group for values with (modulus 10, remainder 3);
    create table public.recommendation_partial_reused_group_p4 partition of public.recommendation_partial_reused_group for values with (modulus 10, remainder 4);
    create table public.recommendation_partial_reused_group_p5 partition of public.recommendation_partial_reused_group for values with (modulus 10, remainder 5);
    create table public.recommendation_partial_reused_group_p6 partition of public.recommendation_partial_reused_group for values with (modulus 10, remainder 6);
    create table public.recommendation_partial_reused_group_p7 partition of public.recommendation_partial_reused_group for values with (modulus 10, remainder 7);
    create table public.recommendation_partial_reused_group_p8 partition of public.recommendation_partial_reused_group for values with (modulus 10, remainder 8);
    create table public.recommendation_partial_reused_group_p9 partition of public.recommendation_partial_reused_group for values with (modulus 10, remainder 9);

    --
    -- migrate the data from old table to the new table
    --
    PERFORM public.migrate_data('public', 'recommendation_partial_reused_group_old', 'recommendation_partial_reused_group');

    --
    -- Add the indexes for recommendation_partial_reused_group
    --
    DROP INDEX IF EXISTS public.rec_partial_reused_path_idx;
    CREATE INDEX rec_partial_reused_path_idx ON public.recommendation_partial_reused_group USING btree (path);

    DROP INDEX IF EXISTS public.rec_partial_reuse_rec_path_idx;
    CREATE INDEX rec_partial_reuse_rec_path_idx ON public.recommendation_partial_reused_group USING btree (recommendation_path);

    CREATE INDEX rec_partial_reuse_site_id_idx ON public.recommendation_partial_reused_group USING btree (site_id);

    --
    -- Add foreign key
    --
    ALTER TABLE public.recommendation_partial_reused_group
    ADD CONSTRAINT rec_partial_reuse_foreign FOREIGN KEY (recommendation_path, rec_site_id) REFERENCES public.recommendation(path, site_id);

    --
    -- Drop the old table
    --
    DROP TABLE IF EXISTS public.recommendation_partial_reused_group_old;
    DROP TABLE IF EXISTS public.recommendation_exclusion_old;
    DROP TABLE IF EXISTS public.recommendation_monitor_old;
    DROP TABLE IF EXISTS public.recommendation_old;

    --
    -- Create tables for auto tagging
    -- extends TaggingRuleEntity extends AbstractManagedResource class, hence specify all the columns in AbstractManagedResource here
    --
    CREATE TABLE IF NOT EXISTS public.tagging_rule (
        id CHARACTER VARYING(128),
        displayname CHARACTER VARYING(128),
        flow_src_sites VARCHAR(128)[] DEFAULT '{}',
        flow_dst_sites VARCHAR(128)[] DEFAULT '{}',
        flow_protocol_and_dst_port_pairs VARCHAR(128)[] DEFAULT '{}',
        flow_app_ids INTEGER[] DEFAULT '{}',
        flow_app_guids VARCHAR(64)[] DEFAULT '{}',
        src_tags VARCHAR(128)[] DEFAULT '{}',
        dst_tags VARCHAR(128)[] DEFAULT '{}',
        activated BOOLEAN DEFAULT FALSE NOT NULL,
        revision INTEGER DEFAULT 0 NOT NULL,
        description character varying(1024),
        systemowned boolean,
        createuser CHARACTER VARYING(64),
        createtime BIGINT,
        lastmodifieduser CHARACTER VARYING(64),
        lastmodifiedtime BIGINT,
        PRIMARY KEY(id)
    );

    CREATE TABLE IF NOT EXISTS public.tagging_rule_monitor (
        tagging_rule_id CHARACTER VARYING(128),
        num_tags_assigned INTEGER DEFAULT 0 NOT NULL,
        num_successful_operations INTEGER DEFAULT 0 NOT NULL,
        num_failed_operations INTEGER DEFAULT 0 NOT NULL,
        failed_operation_info_list JSONB DEFAULT '{}'::JSONB,
        PRIMARY KEY(tagging_rule_id),
        CONSTRAINT fk_tagging_rule
            FOREIGN KEY(tagging_rule_id)
                REFERENCES public.tagging_rule(id)
                ON DELETE CASCADE
    );

    RETURN TRUE;
END;
$$
LANGUAGE 'plpgsql';

--
--Create recommendation postgres schema
--
CREATE OR REPLACE FUNCTION public.create_psql_schema()
RETURNS BOOLEAN
AS
$$
DECLARE
    _num integer;
    _tmp BOOLEAN;
BEGIN
    --
    -- Check if recommendation table exists, for Saas we need to create all the recommendation table
    --
    EXECUTE format('SELECT EXISTS(SELECT 1 FROM information_schema.tables WHERE
    table_schema=%L
    AND table_name=%L)', 'public', 'recommendation')
    INTO _tmp;
    --
    -- Create the recommendation table for Saas case
    --
    IF NOT _tmp THEN
        RAISE INFO 'No rec table exists, start to create the rec partition table now';
        PERFORM public.create_partition_table();
        RAISE INFO 'Finish creating the partition table';
    ELSE
        RAISE INFO 'Rec table exists, checking if rec table has partition key';
        EXECUTE format('SELECT count(*) AS partitions FROM pg_catalog.pg_inherits WHERE inhparent=%L::regclass', 'public.recommendation')
        INTO _num;
        --
        -- BEGIN AND END serve as transaction block here,
        -- so other tables' partitions will be created if recommendation table's partitions are created successfully
        -- we only need to check recommendation table's partitions to see if the schema is up to date.
        -- If the table contains partitions, then no need to do anything
        -- If the table does not contain the partition, then this is for the om-prem upgrade case to upgrade the old schema
        --
        IF _num = 0 THEN
            RAISE INFO 'Rec table does not have partition key, upgrade scheme now';
            PERFORM public.create_partition_table();
            RAISE INFO 'Finish upgrading schema';
        END IF;
    END IF ;
    RAISE INFO 'Finish creating all recommendation tables';
    RETURN TRUE;
END;
$$
LANGUAGE 'plpgsql';

SELECT public.create_psql_schema();