import argparse
import json
import sys
import time
import ssl
import os
import logging
from kubernetes import client, config, watch

from task_common import read_task_file
from task_common import post_task
from task_common import await_task_completion
from task_common import await_load_completion

from datetime import datetime, timedelta

FORMAT = '%(asctime)s %(levelname)s: %(message)s'
DATEFMT = '%Y-%m-%d %H:%M:%S'
logging.basicConfig(format=FORMAT, datefmt=DATEFMT, level=logging.DEBUG)


def get_last_successful_job_timestemp(args):
  config.load_incluster_config()
  jobs = client.BatchV1Api().list_namespaced_job(os.getenv("POD_NAMESPACE"))
  datetimes = [job.status.completion_time for job in jobs.items \
    if job.metadata.name.startswith(args.job_type) and job.status.succeeded == 1]
  return sorted(datetimes)[-1]


def get_daily_rollup_interval(args):
  # To preserve data with original granularity for the
  # past day end timestamp for rollup is 1 day before the
  # past day
  # for correlated_flow_viz, we always have the same granularity, no need to preserve
  # original granularity, we accept value passed by user from command line.
  interval_end_timestamp = datetime.today() - timedelta(days=args.guard_interval_days)
  # Read last timestamp when rollup succeeded from success file
  # If file does not exist or is corrupted reset the last success
  # timestamp to 30 days before current time.
  try:
    # remove timezone info, because datetime.today() doesn't generate timezone info
    # and timezone should be consistent in single kubernetes environment
    success_timestamp = get_last_successful_job_timestemp(args).replace(tzinfo=None)
  except Exception as e:
    logging.info("got exception when fetching last completed reindex job timestamp: {0}".format(e))
    success_timestamp = interval_end_timestamp - timedelta(days=30)
    pass
  logging.info("last success timestemp: {0}".format(success_timestamp))
  interval_start_timestamp = min(success_timestamp - timedelta(days=args.guard_interval_days), interval_end_timestamp - timedelta(days=args.num_days))
  time_interval = interval_start_timestamp.strftime('%Y-%m-%d') + \
                  "/" + \
                  interval_end_timestamp.strftime('%Y-%m-%d')
  return interval_end_timestamp, time_interval


def get_hourly_rollup_interval():
  interval_end_timestamp = datetime.today() - timedelta(hours=12)
  interval_start_timestamp = interval_end_timestamp - timedelta(hours=1)
  time_interval = interval_start_timestamp.strftime('%Y-%m-%dT%H:00:00') + \
                  "/" + \
                  interval_end_timestamp.strftime('%Y-%m-%dT%H:00:00')
  return interval_end_timestamp, time_interval


def main():
  parser = argparse.ArgumentParser(description='Post Druid indexing tasks.')
  parser.add_argument('--data-source', '-d', type=str, default='correlated_flow_rec', help='The name of Druid table (dataSource) to be rolled up')
  parser.add_argument('--url', '-u', metavar='url', type=str, default='https://druid-coordinator:8281/', help='Druid Overlord url')
  parser.add_argument('--coordinator-url', type=str, default='https://druid-coordinator:8281/', help='Druid Coordinator url')
  parser.add_argument('--file', '-f', type=str, required=True, help='Query JSON file')
  parser.add_argument('--submit-timeout', type=int, default=120, help='Timeout (in seconds) for submitting tasks')
  parser.add_argument('--complete-timeout', type=int, default=14400, help='Timeout (in seconds) for completing tasks')
  parser.add_argument('--load-timeout', type=int, default=14400, help='Timeout (in seconds) for waiting for tasks to load')
  parser.add_argument('--quiet', '-q', action='store_true', help='Suppress retryable errors')
  parser.add_argument('--user', type=str, default=None, help='Basic auth username')
  parser.add_argument('--password', type=str, default=None, help='Basic auth password')
  parser.add_argument('--granularity', '-g', type=str, required=True, default=None, help='Segment and query granularity')
  parser.add_argument('--num-days', '-n', type=int, required=True, default=None, help='Number of days from when to rollup')
  parser.add_argument('--guard-interval-days', type=int, default=2, help='Number of days back from now when rollup ends (default is 2 days)')
  parser.add_argument('--job-type', type=str, required=True, help='reindex job type, this should the name of the cronjob')
  args = parser.parse_args()

  submit_timeout_at = time.time() + args.submit_timeout
  complete_timeout_at = time.time() + args.complete_timeout

  task_contents = read_task_file(args.file)
  task_json = json.loads(task_contents)

  datasource = args.data_source
  logging.info("Beginning indexing data for {0}".format(datasource))
  task_json["spec"]["ioConfig"]["inputSource"]["dataSource"] = datasource
  task_json["spec"]["dataSchema"]["dataSource"] = datasource

  interval_end_timestamp, time_interval = get_hourly_rollup_interval() if args.granularity == "hour" \
                  else get_daily_rollup_interval(args)
  task_json["spec"]["ioConfig"]["inputSource"]["interval"] = time_interval
  task_json["spec"]["dataSchema"]["granularitySpec"]["queryGranularity"] = args.granularity
  task_json["spec"]["dataSchema"]["granularitySpec"]["segmentGranularity"] = args.granularity
  task_contents = json.dumps(task_json)

  logging.info("task_spec: {0}".format(task_contents))

  # SSL context initialization
  ctx = ssl.create_default_context()
  ctx.load_verify_locations("/etc/nsxi/certs/ca.crt")

  task_id = json.loads(post_task(args, task_contents, submit_timeout_at, ctx))["task"]

  logging.info('\033[1m' + "Task started: " + '\033[0m' + "{0}".format(task_id))
  logging.info('\033[1m' + "Task log:     " + '\033[0m' + "{0}/druid/indexer/v1/task/{1}/log".format(args.url.rstrip("/"), task_id))
  logging.info('\033[1m' + "Task status:  " + '\033[0m' + "{0}/druid/indexer/v1/task/{1}/status".format(args.url.rstrip("/"), task_id))

  task_status = await_task_completion(args, task_id, complete_timeout_at, ctx)
  logging.info("Task finished with status: {0}".format(task_status))
  if task_status != 'SUCCESS':
    sys.exit(1)

  logging.info("Completed indexing data for ({0}). Now loading indexed data onto the cluster...".format(datasource))
  load_timeout_at = time.time() + args.load_timeout
  await_load_completion(args, datasource, load_timeout_at, ctx)

try:
  main()
except KeyboardInterrupt:
  sys.exit(1)
